@@ -15,7 +15,7 @@
  *
  */
 /*
- *  Copyright (C) 2000-2016 Thomas Habets <thomas@habets.se>
+ *  Copyright (C) 2000-2017 Thomas Habets <thomas@habets.se>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -304,11 +304,54 @@ drop_capabilities()
 #endif
 }
 
+/**
+ * Get GID of input (handle both name and number) or die.
+ */
+static gid_t
+must_get_group(const char* ident)
+{
+        // Special case empty string, because strtol.
+        int saved_errno;
+        if (*ident) {
+                // First try it as a name.
+                {
+                        struct group* gr;
+                        errno = 0;
+                        if (gr = getgrnam(ident)) {
+                                return gr->gr_gid;
+                        }
+                        saved_errno = errno;
+                }
+
+                // Not a name. Try it as an integer.
+                {
+                        char* endp = NULL;
+                        gid_t r = strtol(ident, &endp, 0);
+                        if (!*endp) {
+                                return r;
+                        }
+                }
+        }
+
+        if (saved_errno != 0) {
+                fprintf(stderr,
+                        "arping: %s not a number and getgrnam(%s): %s\n",
+                        ident, ident, strerror(saved_errno));
+        } else {
+                // If group was supplied, then not
+                // existing is fatal error too.
+                fprintf(stderr,
+                        "arping: %s is not a number or group\n",
+                        ident);
+        }
+        exit(1);
+}
+
 /**
  * drop all privileges.
  */
 static void
-drop_privileges()
+drop_privileges(const char* drop_group)
 {
         // Need to get uid/gid of 'nobody' before chroot().
         const char* drop_user = "nobody";
@@ -318,14 +361,23 @@ drop_privileges()
         gid_t gid = 0;
         if (!(pw = getpwnam(drop_user))) {
                 if (verbose) {
-                        printf("arping: getpwnam(%s): %s\n",
-                               drop_user, strerror(errno));
+                        if (errno != 0) {
+                                printf("arping: getpwnam(%s): %s\n",
+                                       drop_user, strerror(errno));
+                        } else {
+                                printf("arping: getpwnam(%s): unknown user\n",
+                                       drop_user);
+                        }
                 }
-                return;
         } else {
                 uid = pw->pw_uid;
                 gid = pw->pw_gid;
         }
+
+        // If group is supplied, use that gid instead.
+        if (drop_group != NULL) {
+                gid = must_get_group(drop_group);
+        }
         drop_fs_root();
         drop_uid(uid, gid);
         drop_capabilities();
@@ -602,6 +654,8 @@ extended_usage()
                "    -e     Like -a but beep when there is no reply.\n"
 	       "    -F     Don't try to be smart about the interface name.  (even  if  this\n"
 	       "           switch is not given, -i overrides smartness)\n"
+               "    -g group\n"
+               "           setgid() to this group instead of the nobody group.\n"
 	       "    -h     Displays a help message and exits.\n"
 	       "    -i interface\n"
 	       "           Use the specified interface.\n"
@@ -638,8 +692,9 @@ extended_usage()
 	       "    -U     Send unsolicited ARP.\n"
 	       "    -v     Verbose output. Use twice for more messages.\n"
                "    -V num 802.1Q tag to add. Defaults to no VLAN tag.\n"
-               "    -w     Time to wait between pings, in microseconds.\n"
-               "    -W     Same as -w, but in floating point seconds.\n");
+               "    -w sec Specify a timeout before ping exits regardless of how"
+               " many\npackets have been sent or received.\n"
+               "    -W sec Time to wait between pings.\n");
         printf("Report bugs to: thomas@habets.se\n"
                "Arping home page: <http://www.habets.pp.se/synscan/>\n"
                "Development repo: http://github.com/ThomasHabets/arping\n");
@@ -653,14 +708,15 @@ standard_usage()
 {
 	printf("ARPing %s, by Thomas Habets <thomas@habets.se>\n",
 	       version);
-        printf("usage: arping [ -0aAbdDeFpPqrRuUv ] [ -w <us> ] "
+        printf("usage: arping [ -0aAbdDeFpPqrRuUv ] [ -w <sec> ] "
                "[ -W <sec> ] "
                "[ -S <host/ip> ]\n"
                "              "
                "[ -T <host/ip ] "
                "[ -s <MAC> ] [ -t <MAC> ] [ -c <count> ]\n"
                "              "
-               "[ -C <count> ] [ -i <interface> ] [ -m <type> ]\n"
+               "[ -C <count> ] [ -i <interface> ] [ -m <type> ]"
+               " [ -g <group> ]\n"
                "              "
                "[ -V <vlan> ] [ -Q <priority> ] "
                "<host/ip/MAC | -B>\n");
@@ -782,6 +838,30 @@ timespec2dbl(const struct timespec *tv)
         return tv->tv_sec + (double)tv->tv_nsec / 1000000000;
 }
 
+/**
+ * return number of microseconds to wait for packets.
+ */
+static uint32_t
+wait_time(double deadline, uint32_t packetwait)
+{
+        struct timespec ts;
+
+        // If deadline not specified, then don't use it.
+        if (deadline < 0) {
+                return packetwait;
+        }
+
+        getclock(&ts);
+        const double max_wait = deadline - timespec2dbl(&ts);
+        if (max_wait < 0) {
+                return 0;
+        }
+        if (max_wait > packetwait / 1000000) {
+                return packetwait;
+        }
+        return max_wait * 1000000;
+}
+
 /**
  * max size of buffer is intsize + 1 + intsize + 4 = 70 bytes or so
  *
@@ -839,7 +919,7 @@ pingmac_send(uint16_t id, uint16_t seq)
         // Without this padding some systems (e.g. Raspberry Pi 3
         // wireless interface) failed. dmesg said:
         //   arping: packet size is too short (42 <= 50)
-        const char padding[16] = {0};
+        const uint8_t padding[16] = {0};
 
 	int c;
 
@@ -930,7 +1010,7 @@ pingip_send()
         // Without this padding some systems (e.g. Raspberry Pi 3
         // wireless interface) failed. dmesg said:
         //   arping: packet size is too short (42 <= 50)
-        const char padding[16] = {0};
+        const uint8_t padding[16] = {0};
 
 	if (-1 == (arp = libnet_build_arp(ARPHRD_ETHER,
 					  ETHERTYPE_IP,
@@ -985,7 +1065,7 @@ pingip_send()
                        (long)lastpacketsent.tv_nsec);
 	}
 	if (-1 == libnet_write(libnet)) {
-		fprintf(stderr, "arping: libnet_write(): %s\n", 
+		fprintf(stderr, "arping: libnet_write(): %s\n",
 			libnet_geterror(libnet));
 		sigint(0);
 	}
@@ -1288,6 +1368,11 @@ ping_recv(pcap_t *pcap, uint32_t packetwait, pcap_handler func)
        fixup_timespec(&endtime);
 
        fd = pcap_get_selectable_fd(pcap);
+       if (fd == -1) {
+               fprintf(stderr, "arping: pcap_get_selectable_fd()=-1: %s\n",
+                       pcap_geterr(pcap));
+               exit(1);
+       }
        old_received = numrecvd;
 
        for (;!done;) {
@@ -1429,10 +1514,11 @@ arping_main(int argc, char **argv)
 	int dstip_given = 0;
         const char *srcmac_opt = NULL;
         const char *dstmac_opt = NULL;
-	const char *ifname = NULL;
+	const char *ifname = NULL;      // -i/-I
         int opt_B = 0;
         int opt_T = 0;
         int opt_U = 0;
+        const char* drop_group = NULL;  // -g
         const char *parm; // First argument, meaning the target IP.
 	int c;
 	unsigned int maxcount = -1;
@@ -1440,7 +1526,8 @@ arping_main(int argc, char **argv)
 	struct bpf_program bp;
 	pcap_t *pcap;
 	enum { NONE, PINGMAC, PINGIP } mode = NONE;
-	unsigned int packetwait = 1000000;
+	unsigned int packetwait = 1000000; // Default one second.
+        double deadline = -1;
         char bpf_filter[64];
         ebuf[0] = 0;
 
@@ -1457,7 +1544,7 @@ arping_main(int argc, char **argv)
 	memcpy(dstmac, ethxmas, ETH_ALEN);
 
         while (EOF != (c = getopt(argc, argv,
-                                  "0aAbBC:c:dDeFhi:I:m:pPqQ:rRs:S:t:T:uUvV:w:W:"))) {
+                                  "0aAbBC:c:dDeFg:hi:I:m:pPqQ:rRs:S:t:T:uUvV:w:W:"))) {
 		switch(c) {
 		case '0':
 			srcip_opt = "0.0.0.0";
@@ -1496,6 +1583,9 @@ arping_main(int argc, char **argv)
 			break;
 		case 'h':
 			usage(0);
+                case 'g':
+                        drop_group = optarg;
+                        break;
 		case 'i':
 			if (strchr(optarg, ':')) {
 				fprintf(stderr, "arping: If you're trying to "
@@ -1574,7 +1664,7 @@ arping_main(int argc, char **argv)
                         }
                         break;
 		case 'w':
-			packetwait = (unsigned)atoi(optarg);
+                        deadline = atof(optarg);
 			break;
                 case 'W':
                         packetwait = (unsigned)(1000000.0 * atof(optarg));
@@ -1592,7 +1682,7 @@ arping_main(int argc, char **argv)
         }
 
         if (((mode == PINGIP) && opt_T)
-            || (mode == PINGMAC) && (opt_B || dstmac_opt || opt_U)) {
+            || ((mode == PINGMAC) && (opt_B || dstmac_opt || opt_U))) {
                 fprintf(stderr, "arping: -T can only be used to ping MAC"
                         " and -BtU only to ping IPs");
                 exit(1);
@@ -1650,9 +1740,14 @@ arping_main(int argc, char **argv)
         if (verbose > 1) {
 #if HAVE_CLOCK_MONOTONIC
                 struct timespec ts;
-                clock_getres(CLOCK_MONOTONIC, &ts);
-                printf("arping: clock_getres() = %ld %ld\n",
-                       (long)ts.tv_sec, (long)ts.tv_nsec);
+                if (clock_getres(CLOCK_MONOTONIC, &ts)) {
+                        fprintf(stderr,
+                                "arping: clock_getres(CLOCK_MONOTONIC, ...): %s\n",
+                                strerror(errno));
+                } else {
+                        printf("arping: clock_getres() = %lds %ldns\n",
+                               (long)ts.tv_sec, (long)ts.tv_nsec);
+                }
 #else
                 printf("arping: Using gettimeofday() for time measurements\n");
 #endif
@@ -1694,7 +1789,7 @@ arping_main(int argc, char **argv)
                         do_libnet_init(ifname, 0);
 			parm = strdup(libnet_addr2name4(dstip,0));
 			if (!parm) {
-				fprintf(stderr, "arping: out of mem\n");
+				fprintf(stderr, "arping: out of memory\n");
 				exit(1);
 			}
 		}
@@ -1812,7 +1907,7 @@ arping_main(int argc, char **argv)
                 fprintf(stderr, "arping: pcap_open_live(): %s\n", ebuf);
 		exit(1);
 	}
-        drop_privileges();
+        drop_privileges(drop_group);
 	if (pcap_setnonblock(pcap, 1, ebuf)) {
                 strip_newline(ebuf);
 		fprintf(stderr, "arping: pcap_set_nonblock(): %s\n", ebuf);
@@ -1845,8 +1940,8 @@ arping_main(int argc, char **argv)
                         snprintf(bpf_filter, sizeof(bpf_filter), "arp");
                 }
                 if (-1 == pcap_compile(pcap, &bp, bpf_filter, 0, -1)) {
-                        fprintf(stderr, "arping: pcap_compile(%s): error\n",
-                                bpf_filter);
+                        fprintf(stderr, "arping: pcap_compile(%s): %s\n",
+                                bpf_filter, pcap_geterr(pcap));
 			exit(1);
 		}
 	} else { /* ping mac */
@@ -1858,13 +1953,14 @@ arping_main(int argc, char **argv)
                         snprintf(bpf_filter, sizeof(bpf_filter), "icmp");
                 }
                 if (-1 == pcap_compile(pcap, &bp, bpf_filter, 0,-1)) {
-                        fprintf(stderr, "arping: pcap_compile(%s): error\n",
-                                bpf_filter);
+                        fprintf(stderr, "arping: pcap_compile(%s): %s\n",
+                                bpf_filter, pcap_geterr(pcap));
 			exit(1);
 		}
 	}
 	if (-1 == pcap_setfilter(pcap, &bp)) {
-		fprintf(stderr, "arping: pcap_setfilter(): error\n");
+                fprintf(stderr, "arping: pcap_setfilter(): %s\n",
+                        pcap_geterr(pcap));
 		exit(1);
 	}
 
@@ -1908,14 +2004,22 @@ arping_main(int argc, char **argv)
 	/*
 	 * let's roll
 	 */
+        if (deadline > 0) {
+                struct timespec ts;
+                getclock(&ts);
+                deadline += timespec2dbl(&ts);
+        }
 	if (mode == PINGIP) {
 		unsigned int c;
                 unsigned int r;
 		for (c = 0; c < maxcount && !time_to_die; c++) {
 			pingip_send();
                         r = numrecvd;
-			ping_recv(pcap,packetwait,
-				  (pcap_handler)pingip_recv);
+                        const uint32_t w = wait_time(deadline, packetwait);
+                        if (w == 0) {
+                                break;
+                        }
+                        ping_recv(pcap, w, (pcap_handler)pingip_recv);
 		}
 	} else { /* PINGMAC */
 		unsigned int c;
@@ -1923,8 +2027,11 @@ arping_main(int argc, char **argv)
 		for (c = 0; c < maxcount && !time_to_die; c++) {
 			pingmac_send(rand(), c);
                         r = numrecvd;
-			ping_recv(pcap,packetwait,
-				  (pcap_handler)pingmac_recv);
+                        const uint32_t w = wait_time(deadline, packetwait);
+                        if (w == 0) {
+                                break;
+                        }
+                        ping_recv(pcap, w,  (pcap_handler)pingmac_recv);
 		}
 	}
         if (display == DOT) {
