@@ -1,6 +1,6 @@
 /* arping/src/arping.h
  *
- *  Copyright (C) 2000-2011 Thomas Habets <thomas@habets.se>
+ *  Copyright (C) 2000-2015 Thomas Habets <thomas@habets.se>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -25,9 +25,24 @@
 #include <inttypes.h>
 #endif
 
+/* Forward declarations */
+struct pcap_pkthdr;
+
 extern int verbose;
+extern uint32_t srcip;
+extern uint32_t dstip;
+extern unsigned int numrecvd;
+extern unsigned int numsent;
+
+const char *
+arping_lookupdev(uint32_t srcip, uint32_t dstip, char *ebuf);
 void do_signal_init();
 void do_libnet_init(const char *ifname, int recursive);
 void sigint(int);
 const char *arping_lookupdev_default(uint32_t srcip, uint32_t dstip,
 				     char *ebuf);
+int arping_main(int argc, char **argv);
+
+
+void pingip_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet);
+void pingmac_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet);
